#
# This example shows how a specialized meta class can be used to
# construct a class system.  This one happens to work the same as the
# one I have coded in C.  Without the C code in place to start with,
# the Python code would have to create a handful of "throw away"
# classes.  Because the C code is available, the system can be
# expressed easily in terms of itself.
#

import object, types

class base(object.base):
    #
    # __init__ gets called when an instance of base is being initialized.
    #
    def __init__(self, *args, **kw):
        print 'base.__init__', self, args, kw



    #----------
    #
    # This class statement defines the class of the base class.
    # As you would expect and the sytax implies, it can be
    # referred to as base.__class__.  It's repr function
    # will tell you that its name is 'base_class'.
    #
    class __class__:

        #
        # __init__ gets called on a newly created instances of
        # base_class or one of it's sub-classes.
        #
        def __init__(self, name, bases=None, dict=None):
            print 'base_class.__init__', name, bases, dict


        #
        # __new__ gets called as part of the instantiation process of
        # base.  It is expected to return an uninitialized instance of
        # the class indicated by 'self'.  This should be base or one
        # of it's sub-classes.
        #
        def __new__(self, *args, **kw):
            super = object.base.__class__._
            i = super.__new__(self)
            print 'base_class.__new__', self, args, kw, '->', i
            return i

            
        #
        # The __call__ method gets invoked whenever base gets "called"
        # it is expected to return a fully initialized instance of
        # base or one of it's sub-classes
        #
        def __call__(self, *args, **kw):
            print 'base_class.__call__', args, kw
            i = apply(self.__new__, args, kw)
            try:
                init = i.__init__
            except AttributeError:
                pass
            else:
		apply(init, args, kw)
            return i
            
            
        #-----------
        #
        # The following class defines the class of base.__class__,
        # which of course can be refered to as
        # base.__class__.__class__.  Its repr function will tell you
        # that it's name is base_class_class.
        #
            
        class __class__:
            #
            # __init__ will get called on each new instance of 
            # base.__class__.__class__.
            #
            def __init__(self, name, bases=None, dict=None):
                print 'base_class_class.__init__', self, name, bases, \
                      dict and dict.keys()

            #
            # __new__ will get called when the function call operator
            # is applied to an instance of base.__class__.  This happens
            # internally when ever base.__class__ is subclassed.
            #
            # All class creation (not instance creation, see above
            # for that) in the 'base' hierarchy will go through
            # here unless a sub-class overrides it.
            # 
            def __new__(self, name, bases=None, dict=None):
                print 'base_class_class.__new__', self, name, bases, \
                      dict and dict.keys()
                cl = self
                if dict and dict.has_key('__class__'):
                    cl = dict['__class__']
                    if type(cl) is types.ClassType and bases:
                        base = bases[0]
                        if type(base) is not type(self):
                            cl = None
                        else:
                            meta_class = base.__class__.__class__
                            meta_bases = (base.__class__,)
                            meta_name = "%s_class" % name
                            cl = meta_class(meta_name, meta_bases, cl.__dict__)
                if type(cl) is not type(self):
                    raise TypeError, "invalid class for this context"

                super = object.base.__class__.__class__._
                c= super.__new__(cl, name, bases, dict)
                return c

            #
            #
            #
            def __call__(self, *args, **kw):
                print 'base_class_class.__call__', args, kw
                i = apply(self.__new__, args, kw)
                try:
                    init = i.__init__
                except AttributeError:
                    pass
                else:
		    apply(init, args, kw)
                return i
            


#
# Base is what we have been working for.  It is a class that can be
# specialized. As is stands, it doesn't do anything.  But, whenever
# you derive a new class from it, you will also be deriving a new meta
# class.  The new meta class makes your new class act like a real
# object.  it will inherit behavior from its base class.  Adding new
# class level methods is as simple as writing an instance method.
#

class foo(base):
    
    def __init__(self):
        print 'foo.__init__', self
        super = base._
        super.__init__(self)

    class __class__:
        a = 24
        def __init__(cl, *args, **kw):
            print 'foo_class.__init__', cl, args, kw

        def __new__(cl, *args, **kw):
            super = base.__class__._
            print 'foo_class.__new__', cl, args, kw
            return super.__new__(cl)

        def spam(cl, x):
            cl.a = x
            cl._.a = x / 2;
            print cl.a, cl._.a

            

print foo
print foo.__class__


class bar(foo):
    a = 10
    def __init__(self):
        super = foo._
        print 'bar.__init__', self
        print super.__init__.func_code
        super.__init__(self)
        
    def __new__(cl, *args, **kw):
        super = foo.__class__._
        print 'bar_class.__new__', cl, args, kw
        return super.__new__(cl)

	
class box(bar):
    a = 10
    def __init__(self):
        super = foo._
        print 'bar.__init__', self
        print super.__init__.func_code
        super.__init__(self)
        
    def __new__(cl, *args, **kw):
        super = foo.__class__._
        print 'bar_class.__new__', cl, args, kw
        return super.__new__(cl)
        
