/***********************************************************
 Copyright (c) Donald Beaudry, 1996.

 This program is freely distributable without licensing fees
 and is provided without guarantee or warrantee expressed or
 implied.
 
***************************************************************/
/* $Id: objectmodule.c,v 1.1 1996/12/02 19:16:05 donb Exp donb $ */

/* Module: object -- an alternate class system */

#include "Python.h"
#include "compile.h"
#include "structmember.h"
#include <assert.h>

/* #define TRACING /**/
#ifdef TRACING
int _trace_level_ = 0;

#define TFILE strrchr(__FILE__,'/')+1

#define ENTER(x) \
    static char *_func_name_ = x; \
    fprintf(stderr, "%15s:%4d:%*s%s\n", \
            TFILE, __LINE__, (++_trace_level_)*2, "(", _func_name_)

#define _LEAVE \
    fprintf(stderr, "%15s:%4d:%*s%s\n", \
            TFILE, __LINE__, (_trace_level_--)*2, ")", _func_name_)

#define TFPRINTF \
    fprintf(stderr, "%15s:%4d:%*s", \
            TFILE, __LINE__, _trace_level_*2, " "),fprintf

#define RETURN return _LEAVE,
#define RETURNV do {_LEAVE; return;} while(0)
#define TRACEF(t,x)  TFPRINTF(stderr, "%s="t"\n", #x, x)
#define TRACE(x) TFPRINTF(stderr, "%s\n", x)
#else
#define ENTER(x)
#define RETURN return
#define RETURNV return
#define TRACE(x)
#define TRACEF(t,x)
#define TFPRINTF 0 &&
#endif



static PyObject *BoundMethod_New Py_PROTO((PyObject *func, PyObject *self));

extern DL_IMPORT(PyTypeObject) BoundMethod_Type;

#define BoundMethod_Check(op) ((op)->ob_type == &BoundMethod_Type)


static PyTypeObject MetaType_Type;
static PyTypeObject MetaClass_Type;

static PyObject *MetaClass = 0;
static PyObject *BaseClass = 0;


typedef struct
InstanceAttrAccessorObject
{
  PyObject_HEAD
  struct MetaClassObject *class;	/* The class it is bound to */
} InstanceAttrAccessorObject;


typedef struct MetaClassObject
{
  PyClassObject py;
  struct MetaClassObjectPart
    {
      struct MetaClassObject *class;
      PyObject *dict;
      ternaryfunc instantiate;
      InstanceAttrAccessorObject *accessor;
    } mc;
} MetaClassObject;

#define MetaClass_Check(op) (((PyObject*)(op))->ob_type == &MetaClass_Type)

static PyObject *MetaClass_Call Py_PROTO((MetaClassObject *bm, PyObject *args,
                                          PyObject *kw));
static PyObject *MetaClass_New Py_PROTO((PyObject *class, PyObject *bases,
                                         PyObject *methods, PyObject *name));
     
InstanceAttrAccessorObject *
  InstanceAttrAccessor_New Py_PROTO((MetaClassObject *class));
     


PyObject *
#ifdef HAVE_STDARG_PROTOTYPES
/* VARARGS 2 */
String_Build(char *out_format, char *build_format, ...)
#else
/* VARARGS */
String_Build(va_alist) va_dcl
#endif
{
  va_list va;
  PyObject *args, *retval, *fmt;
#ifdef HAVE_STDARG_PROTOTYPES
  va_start(va, build_format);
#else
  char *build_format;
  char *out_format;
  va_start(va);
  out_format = va_arg(va, char *);
  build_format   = va_arg(va, char *);
#endif

  if (build_format)
    args = Py_VaBuildValue(build_format, va);
  else
    args = PyTuple_New(0);
  
  va_end(va);
  
  if(! args)
    return NULL;

  if (! PyTuple_Check(args))
    {
      PyObject *a;
      
      a=PyTuple_New(1);
      if (! a)
        return NULL;
      
      if (PyTuple_SetItem(a,0,args) == -1)
        return NULL;
      
      args=a;
    }

  fmt = PyString_FromString(out_format);
  
  retval = PyString_Format(fmt, args);
  
  Py_DECREF(args);
  Py_DECREF(fmt);
  
  return retval;
}


/*
 | Invoked when a meta class is sub-classed, results in a new class or
 | meta class.  Depending on the kind of class being sub-classed.
 */

static PyObject *
MetaType_Call(bm, args, kw)
     PyTypeObject *bm;
     PyObject *args;
     PyObject *kw;
{
  PyObject *name;
  PyObject *bases;
  PyObject *methods = NULL;
  PyObject *co;
  MetaClassObject *cl;
  PyObject *base;

  if (!PyArg_ParseTuple(args, "O!O!|O", &PyString_Type, &name,
                        &PyTuple_Type, &bases, &methods))
    return NULL;

  if (PyTuple_Size(bases) < 1)
    {
      PyErr_SetString(PyExc_ValueError, "no bases specified");
      return NULL;
    }

  base = PyTuple_GetItem(bases, 0);
  cl = ((MetaClassObject*)base)->mc.class;

  if (base->ob_type != bm)
    {
      char buffer[200];
      sprintf(buffer, "%.150s expected in bases", bm->tp_name);
      PyErr_SetString(PyExc_TypeError, buffer);
      return NULL;
    }

  co = MetaClass_Call(cl, args, kw);

  return co;
}


/*
 | search the class hierarchy for a value
 */
static PyObject *
class_lookup(cp, name, pclass)
     MetaClassObject *cp;
     PyObject *name;
     MetaClassObject **pclass;
{
  int i, n;
  PyObject *value;

  value = PyDict_GetItem(cp->py.cl_dict, name);
  if (value != NULL)
    {
      if (pclass)
        *pclass = cp;
      return value;
    }
  
  n = PyTuple_Size(cp->py.cl_bases);
  for (i = 0; i < n; i++)
    {
      PyObject *v;
      v = class_lookup((MetaClassObject *) PyTuple_GET_ITEM(cp->py.cl_bases, i),
                       name, pclass);
      if (v != NULL)
        return v;
    }
  
  return NULL;
}


static PyObject *
MetaClass_GetAccessor(op)
     MetaClassObject *op;
{
  /* might need to revive the accessor object */
  if (! op->mc.accessor->ob_refcnt)
    /* class == op it sure better!*/
    Py_INCREF((PyObject*)op->mc.accessor->class);  

  Py_INCREF((PyObject*)op->mc.accessor);
  return (PyObject*) op->mc.accessor;
}


/*
 | The py.cl_dict dictionary of a meta class or base class is there
 | soly to support the instances of the class.  When accessing an
 | attribute through the class object, that dictionary is not
 | considered at all.  Sometimes however, it is desirable to
 | get at the py.cl_dict (method chaining for instance).  For this
 | reason, the this class supports the (pick one:  mapping method
 | or _ attribute).  (note: since the standard python class object
 | never has any attributes associated with it except for those that
 | are there for the instances, there was never a conflict.  But,
 | with theses classes, my_class.__init__ could mean one of two things:
 | the unbound instance method <NULL, my_class.__init__> or the bound
 | instance method <my_class, my_class.__class__.__init__>)  With
 | this system to get the first, you must say my_class['__init__']
 | or my_class._.__init__ (note: the _ attribute will return an object
 | associated with the class that knows how to access the instance
 | dictionary of the class through it's getattr method.))
 */
static PyObject *
MetaClass_GetAttr1(op, oname)
     MetaClassObject *op;
     PyObject *oname;
{
  PyObject *v;
  MetaClassObject *class;
  char *name = PyString_AsString(oname);

  if (name[0] == '_' && name[1] == '\0')
    return MetaClass_GetAccessor((PyObject*) op);

  if (name[0] == '_' && name[1] == '_')
    {
      if (strcmp(name, "__dict__") == 0)
        {
          if (PyEval_GetRestricted())
            {
              PyErr_SetString(PyExc_RuntimeError,
                          "class.__dict__ not accessible in restricted mode");
              return NULL;
            }
          Py_INCREF(op->py.cl_dict);
          return op->py.cl_dict;
        }
      
      if (strcmp(name, "__cl_dict__") == 0)
        {
          if (PyEval_GetRestricted())
            {
              PyErr_SetString(PyExc_RuntimeError,
                          "class.__cl_dict__ not accessible in restricted mode");
              return NULL;
            }
          Py_INCREF(op->mc.dict);
          return op->mc.dict;
        }
      
      if (strcmp(name, "__bases__") == 0)
        {
          Py_INCREF(op->py.cl_bases);
          return op->py.cl_bases;
        }
      
      if (strcmp(name, "__class__") == 0)
        {
          Py_INCREF((PyObject*)op->mc.class);
          return (PyObject*) op->mc.class;
        }
      
      if (strcmp(name, "__name__") == 0)
        {
          if (op->py.cl_name == NULL)
            v = Py_None;
          else
            v = op->py.cl_name;
          Py_INCREF(v);
          return v;
        }
    }

  v = PyDict_GetItem(op->mc.dict, oname);
  if (! v)
    PyErr_Clear();
  else
    {
      Py_INCREF(v);
      return v;
    }

  v = class_lookup(op->mc.class, oname, &class);
  if (v == NULL)
    {
      PyErr_SetString(PyExc_AttributeError, name);
      return NULL;
    }

  if (! PyCallable_Check(v))
    Py_INCREF(v);
  else if (PyMethod_Check(v))
    v = BoundMethod_New(PyMethod_Function(v), (PyObject*)op);
  else
    v = BoundMethod_New(v, (PyObject*)op);

  return v;
}



static PyObject *
MetaClass_GetAttr(inst, name)
     MetaClassObject *inst;
     PyObject *name;
{
  PyObject *func;
  PyObject *res;
  
  res = MetaClass_GetAttr1(inst, name);
  
  if (res == NULL && (func = inst->mc.class->py.cl_getattr) != NULL)
    {
      PyObject *args;
      
      PyErr_Clear();
      args = Py_BuildValue("(OO)", inst, name);
      if (args == NULL)
        return NULL;
      res = PyEval_CallObject(func, args);
      Py_DECREF(args);
    }
  
  return res;
}




static int
MetaClass_SetAttr1(op, oname, v)
     MetaClassObject *op;
     PyObject *oname;
     PyObject *v;
{
  char *name = PyString_AsString(oname);
     
  if (name[0] == '_' && name[1] == '_')
    {
      int n = strlen(name);
      if (name[n-1] == '_' && name[n-2] == '_')
        {
          PyErr_SetString(PyExc_TypeError, "read-only special attribute");
          return -1;
        }
    }
  
  if (PyEval_GetRestricted())
    {
      PyErr_SetString(PyExc_RuntimeError,
                      "classes are read-only in restricted mode");
      return -1;
    }
  
  if (v == NULL)
    {
      int rv = PyDict_DelItem(op->mc.dict, oname);
      if (rv < 0)
        PyErr_SetString(PyExc_AttributeError,
                        "delete non-existing class attribute");
      return rv;
    }
  
  return PyDict_SetItem(op->mc.dict, oname, v);
}


static int
MetaClass_SetAttr(inst, name, v)
     MetaClassObject *inst;
     PyObject *name;
     PyObject *v;
{
  PyObject *func;
  PyObject *args;
  PyObject *res;
  char *sname = PyString_AsString(name);
  
  if (sname[0] == '_' && sname[1] == '_'
      && (strcmp(sname, "__dict__") == 0 || strcmp(sname, "__class__") == 0))
    {
      int n = PyString_Size(name);
      
      if (sname[n-1] == '_' && sname[n-2] == '_')
        {
          PyErr_SetString(PyExc_TypeError, "read-only special attribute");
          return -1;
        }
    }
  
  if (v == NULL)
    func = inst->mc.class->py.cl_delattr;
  else
    func = inst->mc.class->py.cl_setattr;

  if (func == NULL)
    return MetaClass_SetAttr1(inst, name, v);

  if (v == NULL)
    args = Py_BuildValue("(OO)", inst, name);
  else
    args = Py_BuildValue("(OOO)", inst, name, v);

  if (args == NULL)
    return -1;

  res = PyEval_CallObject(func, args);
  Py_DECREF(args);
  if (res == NULL)
    return -1;

  Py_DECREF(res);
  return 0;
}


static void
MetaClass_Dealloc(op)
     MetaClassObject *op;
{
  Py_DECREF(op->mc.dict);
  if (op->mc.class != op)
    Py_DECREF((PyObject*)op->mc.class);

  free(op->mc.accessor);
  
  (*PyClass_Type.tp_dealloc)((PyObject*)op);
}


static PyObject *
MetaClass_Repr(op)
     MetaClassObject *op;
{
  return String_Build("<%s instance %s at %x>", "OOi",
                      ((PyClassObject*)(op->mc.class))->cl_name,
                      op->py.cl_name, op);
}


static PyObject *
MetaClass_Call(bm, args, kw)
     MetaClassObject *bm;
     PyObject *args;
     PyObject *kw;
{
  PyObject *co = NULL;
  PyObject *call;

  ENTER("MetaClass_Call");

  call = PyObject_GetAttrString((PyObject*) bm, "__call__");
  if (call)
    {
      TRACE("found __call__");
      co = PyEval_CallObjectWithKeywords(call, args, kw);
      Py_DECREF(call);
      if (! co)
        RETURN NULL;
    }

  RETURN co;
}



static PyObject *
MetaClass__call__(bm, args, kw)
     MetaClassObject *bm;
     PyObject *args;
     PyObject *kw;
{
  PyObject *co = NULL;
  PyObject *call;

  ENTER("MetaClass__call__");

  if ((PyObject*)bm == Py_None)
    {
      PyErr_SetString(PyExc_TypeError, "call to unbound method");
      return NULL;
    }

  call = PyObject_GetAttrString((PyObject*) bm, "__new__");
  if (call)
    {
      TRACE("found __new__");
      co = PyEval_CallObjectWithKeywords(call, args, kw);
      Py_DECREF(call);
      if (! co)
        {
          TRACE("__new__ failed");
          RETURN NULL;
        }
    }
  else
    {
      MetaClassObject *cl; 
    
      PyErr_Clear();

      TRACE("using default __new__");
        
      cl = (MetaClassObject*)bm->mc.class;
      co = (*cl->mc.instantiate)((PyObject*)bm, args, kw);
      if (! co)
        {
          TRACE("can't instantiate");
          RETURN NULL;
        }
    }

  call = PyObject_GetAttrString(co, "__init__");
  if (! call)
    PyErr_Clear();
  else
    {
      PyObject *ret;

      TRACE("found __init__");
      
      ret = PyEval_CallObjectWithKeywords(call, args, kw);
      Py_DECREF(call);
      if (! ret)
        {
          TRACE("__init__ failed");
          RETURN NULL;
        }
      
      Py_DECREF(ret);
      
      if (ret != Py_None)
        {
          PyErr_SetString(PyExc_TypeError, "__init__() should return None");
          RETURN NULL;
        }
    }
  
  RETURN co;
}



/*
 | Invoked when a call is made to a meta class object, results in new
 | class.
 |
 | If the meta class overrides the __new__ method it becomes its
 | responsibility to construct the new class, if not the default is
 | used.
 */
static PyObject *
MetaClass_Instantiate(bm, args, kw)
     PyObject *bm;
     PyObject *args;
     PyObject *kw;
{
  MetaClassObject *mc;
  PyObject *name;
  PyObject *bases = 0;
  PyObject *methods = 0;
  PyObject *class = 0;

  ENTER("MetaClass_Instantiate");
  
  if (bm == Py_None)
    {
      PyErr_SetString(PyExc_TypeError, "call to unbound method");
      RETURN NULL;
    }

  if (! PyArg_ParseTuple(args, "S|OO", &name, &bases, &methods))
    RETURN NULL;

  if (bases == Py_None)
    bases = NULL;

  if (bases && ! PyTuple_Check(bases))
    {
      PyErr_SetString(PyExc_TypeError, "bases must be tuple or none");
      return NULL;
    }

  class = bm;

  if (methods && PyDict_Check(methods))
    {
      class = PyDict_GetItemString(methods, "__class__");
      if (! class)
        {
          PyErr_Clear();
          class = bm;
        }
      else if (PyClass_Check(class) && bases && PyTuple_Size(bases))
        {
          PyClassObject *co = (PyClassObject*) class;
          MetaClassObject *base = (MetaClassObject*)PyTuple_GET_ITEM(bases, 0);
          if (! MetaClass_Check(base))
            class = NULL;
          else
            {
              PyObject *meta_class =
                (PyObject*)((MetaClassObject*)base->mc.class)->mc.class;
              PyObject *meta_bases = Py_BuildValue("(O)", base->mc.class);
              PyObject *meta_name = String_Build("%s_class", "O", name);
              PyObject *meta_args;
              meta_args = Py_BuildValue("(OOO)", meta_name,
                                        meta_bases, co->cl_dict);

              TRACE("found internal class");
              
              class = MetaClass__call__(meta_class, meta_args, NULL);
              
              Py_DECREF(meta_args);
              Py_DECREF(meta_bases);
              Py_DECREF(meta_name);
              if (! class)
                RETURN NULL;
            }
        }
      else if (! MetaClass_Check(class))
        class = NULL;
    }

  if (! class)
    {
      PyErr_SetString(PyExc_TypeError, "invalid __class__ for this context");
      RETURN NULL;
    }

  mc = (MetaClassObject *) MetaClass_New(class, bases, methods, name);
  if (! mc)
    RETURN NULL;

  /* XXX should merge the keywords with 'methods' */
  if (!methods && kw)
    {
      Py_DECREF(mc->mc.dict);
      Py_INCREF(kw);
      mc->mc.dict = kw;
    }

  RETURN (PyObject*)mc;
}


/*
 | Invoked when a class is called, results in an instance.
 */
static PyObject *
BaseClass_Instantiate(bm, args, kw)
     PyObject *bm;
     PyObject *args;
     PyObject *kw;
{
  PyInstanceObject *inst;

  ENTER("BaseClass_Instantiate");
  
  if (bm == Py_None)
    {
      PyErr_SetString(PyExc_TypeError, "call to unbound method");
      RETURN NULL;
    }
  
  inst = PyObject_NEW(PyInstanceObject, &PyInstance_Type);
  if (inst == NULL)
    RETURN NULL;
  
  Py_INCREF(bm);
  inst->in_class = (PyClassObject *)bm;
  if (! kw)
    inst->in_dict = PyDict_New();
  else
    {
      Py_INCREF(kw);
      inst->in_dict = kw;
    }

  RETURN (PyObject*) inst;
}


/*
 | As with any other object, when creating a metaclass we are
 | instantiating a class.  In this case it is called 'class'.  If
 | 'class' is null we are in bootstrap mode and are creating the first
 | metaclass.  For any given metaclass, there are two ways that it can
 | be used to create another class.  First, it can be sub-classed.
 | Second it can be instantiated (called).
 |
 | When sub-classing at least one base classes must be specified and
 | will match 'class'.  The class of the new class will be the same as
 | the class of 'class'.  Put another way, a sub-class of a class is
 | another instance of it's metaclass.
 |
 | There is one root meta class that is the top of the class
 | hierarchy.  I will refer to it as the MetaClass.  It is an instance
 | of itself and has no base classes.  From it, two different kinds of
 | meta classes can be created: those that create base classes and
 | those that create other meta classes.  Deriving a new class from
 | the MetaClass will create a meta class "factory".  That is, a new
 | meta class that creates more meta classes. Instantiation of the
 | MetaClass will create a new class factory.  That is, a meta class
 | that creates new base classes.  Note that both are called meta
 | classes because they both create instances that are classes.
 |
 | A base class is a class that when instantiated will create an
 | instance.  An instance is not a class.
 | 
 | Either kind of meta class can be derived from further to yield a
 | more specialized meta class.  The more specialized meta class
 | will still generate the same kind of object, either base class
 | or meta class as its super class does.
 |
 | XXXX (pick up here)
 | When a meta class is instantiated, a base class list is allowed,
 | but not required.  The class of the new object is that of the meta
 | class which was instantiated (the object you called).  When a class
 | is derived from, a new instance of the class's class is created.
 | Deriving from a class is therefore and operation on the class's
 | meta class.
 | 
 | When a base class is sub-classed, a new base class is created.
 | When a base class is instantiated, an instanceobject is created.
 |      
 */  

static PyObject *
MetaClass_New(class, bases, methods, name)
     PyObject *class;
     PyObject *bases;
     PyObject *methods;
     PyObject *name;
{
  MetaClassObject *co;
  PyObject *type;
  MetaClassObject *base;

  if (bases == Py_None)
    bases = NULL;

  if (bases && ! PyTuple_Check(bases))
    {
      PyErr_SetString(PyExc_TypeError,
           "meta class instantiation requires a tuple of base classes or None");
      return NULL;
    }

  if (methods && methods != Py_None && ! PyDict_Check(methods))
    {
      PyErr_SetString(PyExc_TypeError,
           "meta class instantiation requires a method dictionary or None");
      return NULL;
    }

  if (bases && PyTuple_Size(bases) > 0)
    base = (MetaClassObject*)PyTuple_GET_ITEM(bases, 0);
  else
    base = NULL;

  if (base && !MetaClass_Check(base))
    {
      PyErr_SetString(PyExc_TypeError, "meta class expected in bases");
      return NULL;
    }
  

  /* XXX need more type checks here. */

  if (methods == NULL || methods == Py_None)
    methods = PyDict_New();
  else
    Py_INCREF(methods);

  co = (MetaClassObject*) PyClass_New(bases, methods, name);

  Py_DECREF(methods);

  co = realloc(co, sizeof (MetaClassObject));

  co->py.ob_type = &MetaClass_Type;
  Py_INCREF(co->py.ob_type);
  
  if (base)
    co->mc.instantiate = base->mc.instantiate;
  else
    co->mc.instantiate = 0;

  Py_XINCREF(class);
  co->mc.class = (MetaClassObject *)class;

  co->mc.dict = PyDict_New();
  co->mc.accessor = NULL;
  co->mc.accessor = InstanceAttrAccessor_New(co);

  return (PyObject*) co;
}


/*--------------------------------
 |
 */


/* InstanceAttrAccessor methods */

/*
 | XXX I'm pretty sure that GetAttr and SetAttr should be invoking
 | methods from the MetaClassPart.  A class that does not want this
 | behavior should not have to create a new InstanceAttrAccessor type
 | to change it.
 */

static PyObject *
InstanceAttrAccessor_GetAttr(ia, name)
     InstanceAttrAccessorObject *ia;
     PyObject *name;
{
  PyObject *v;

  v = PyClass_Type.tp_getattro((PyObject*)ia->class, name);
  if (! v)
    return NULL;
  
  if (PyMethod_Check(v))
    {
      PyObject *f;

      f = PyMethod_Function(v);
      Py_INCREF(f);
      Py_DECREF(v);
      return f;
    }

  if (PyCFunction_Check(v))
    return BoundMethod_New(v, NULL);

  return v;
}


static int
InstanceAttrAccessor_SetAttr(ia, name, v)
     InstanceAttrAccessorObject *ia;
     PyObject *name;
     PyObject *v;
{
  return PyClass_Type.tp_setattro((PyObject*)ia->class, name, v);
}


static void
InstanceAttrAccessor_Dealloc(ia)
     InstanceAttrAccessorObject *ia;
{
  Py_DECREF((PyObject*)ia->class);

  /* note: the memory is not freed here.  It is up to the class to do that.
   * By the time this line gets executed, the memory might already be gone.
   */
}


static int
InstanceAttrAccessor_Compare(a, b)
     InstanceAttrAccessorObject *a, *b;
{
  return PyObject_Compare((PyObject*)a->class, (PyObject*)b->class);
}


static PyObject *
InstanceAttrAccessor_Repr(a)
     InstanceAttrAccessorObject *a;
{
  return String_Build("<class accessor %s>", "O", a->class);
}


static long
InstanceAttrAccessor_Hash(a)
     InstanceAttrAccessorObject *a;
{
  return (long)a ^ (long)a->class;
}


PyTypeObject
InstanceAttrAccessortype =
{
  PyObject_HEAD_INIT(&PyType_Type)
    0,
    "instance attribute accessor",
    sizeof(InstanceAttrAccessorObject),
    0,
    (destructor)InstanceAttrAccessor_Dealloc, /*tp_dealloc*/
    0,                          /*tp_print*/
    0,
    0,
    (cmpfunc)InstanceAttrAccessor_Compare, /*tp_compare*/
    (reprfunc)InstanceAttrAccessor_Repr, /*tp_repr*/
    0,                          /*tp_as_number*/
    0,                          /*tp_as_sequence*/
    0,                          /*tp_as_mapping*/
    (hashfunc)InstanceAttrAccessor_Hash, /*tp_hash*/
    0,                          /*tp_call*/
    0,                          /* tp_str */
    (getattrofunc)InstanceAttrAccessor_GetAttr, /*tp_getattro*/
    (setattrofunc)InstanceAttrAccessor_SetAttr, /*tp_setattro*/
};


/*
 | This object plays some reference counting games.
 |
 | The idea is that since there is only one per class, it will be
 | created when the class is created.  But it will be immediately
 | DECREFed.  Its destructor will not release the memory however.
 | That will happen in the class's dealloc routine.  Instead it will
 | just decref the class.  The class getattr routine will revive the
 | object by incrementing the refcount, and its own if needed.
 */

InstanceAttrAccessorObject *
InstanceAttrAccessor_New(class)
     MetaClassObject *class;
{
  InstanceAttrAccessorObject *ia;

  ia = PyObject_NEW(InstanceAttrAccessorObject, &InstanceAttrAccessortype);
  if (ia == NULL)
    return NULL;
  
  Py_INCREF((PyObject*)class);
  ia->class = class;

  /* won't get freed */
  Py_DECREF(ia);
  
  return ia;
}



static void
MetaClass_InstallMethods(mc, methods)
     PyObject *mc;
     PyMethodDef *methods;
{
  PyObject *d;
  MetaClassObject *cl = (MetaClassObject*) mc;
  PyMethodDef *ml;
  PyObject *v;
  
  d = cl->py.cl_dict;
  
  for (ml = methods; ml->ml_name != NULL; ml++)
    {
      v = PyCFunction_New(ml, Py_None);
      if (v == NULL || PyDict_SetItemString(d, ml->ml_name, v) != 0)
        {
          fprintf(stderr, "initializing class: %s\n",
                  PyString_AsString(cl->py.cl_name));
          Py_FatalError("can't initialize module");
        }
      Py_DECREF(v);
    }
}


PyMethodDef
MetaClass_Methods[] =
{
  {"__call__", (PyCFunction)MetaClass__call__, 1},
  {"__new__", (PyCFunction)MetaClass_Instantiate, 1},

  {NULL},
};



PyMethodDef
BaseClass_Methods[] =
{
  {"__new__", (PyCFunction)BaseClass_Instantiate, 1},

  {NULL},
};


PyMethodDef
Base_Methods[] =
{
  {NULL},
};



/*--------------------------------
 |
 */

typedef struct
BoundMethodObject
{
  PyObject_HEAD
  PyObject *bm_func;	/* The callable object implementing the method */
  PyObject *bm_env;	/* The instance it is bound to, or NULL */
} BoundMethodObject;


#define OFF(x) offsetof(BoundMethodObject, x)

static struct memberlist
BoundMethod_MemberList[] =
{
  {"bm_func",	T_OBJECT,	OFF(bm_func)},
  {"bm_env",	T_OBJECT,	OFF(bm_env)},

  {NULL}	/* Sentinel */
};


static PyObject *
BoundMethod_GetAttr(im, name)
     BoundMethodObject *im;
     char *name;
{
  if (name[0] != '_' && PyEval_GetRestricted())
    {
      PyErr_SetString(PyExc_RuntimeError,
                    "bound-method attributes not accessible in restricted mode");
      return NULL;
    }
  
  return PyMember_Get((char *)im, BoundMethod_MemberList, name);
}


static PyObject *
BoundMethod_Call(bm, arg, kw)
     BoundMethodObject *bm;
     PyObject *arg;
     PyObject *kw;
{
  PyObject *self = bm->bm_env;
  PyObject *func = bm->bm_func;
  PyObject *newarg = NULL;
  PyObject *res;
  int ac;
  int i;
  PyObject *v;

  /* XXX this is bogus (but fast and cheap). */
  if (PyCFunction_Check(func) && PyCFunction_GetSelf(func) == Py_None)
    {
      PyObject *new_args;
      
      if (self)
        return (*((ternaryfunc)PyCFunction_GetFunction(func)))(self, arg, kw);

      if (PyTuple_Size(arg) < 1)
        {
          PyErr_SetString(PyExc_TypeError, "not enough arguments");
          return NULL;
        }
      
      self = PyTuple_GET_ITEM(arg, 0);
      new_args = PyTuple_GetSlice(arg, 1, PyTuple_Size(arg));
      res = (*((ternaryfunc)PyCFunction_GetFunction(func)))(self, new_args, kw);
      
      Py_DECREF(new_args);

      return res;
    }

  ac = PyTuple_Size(arg);
  newarg = PyTuple_New(ac + 1);
  if (newarg == NULL)
    return NULL;
  
  Py_INCREF(self);
  PyTuple_SET_ITEM(newarg, 0, self);

  for (i = 0; i < ac; i++)
    {
      v = PyTuple_GET_ITEM(arg, i);
      Py_INCREF(v);
      PyTuple_SET_ITEM(newarg, i+1, v);
    }

  res = PyEval_CallObjectWithKeywords(func, newarg, kw);
  
  Py_DECREF(newarg);

  return res;
}



static void
BoundMethod_Dealloc(im)
     BoundMethodObject *im;
{
  Py_DECREF(im->bm_func);
  Py_XDECREF(im->bm_env);
  free((ANY *)im);
}


static int
BoundMethod_Compare(a, b)
     BoundMethodObject *a, *b;
{
  if (a->bm_env != b->bm_env)
    return (a->bm_env < b->bm_env) ? -1 : 1;

  return PyObject_Compare(a->bm_func, b->bm_func);
}


static PyObject *
BoundMethod_Repr(a)
     BoundMethodObject *a;
{
  if (a->bm_env)
    return String_Build("<binding (%s, <%s %x>)>", "Osi",
                        a->bm_func, a->bm_env->ob_type->tp_name, a->bm_env);
  else
    return String_Build("<binding (%s, <NULL>)>", "O", a->bm_func);
}


PyTypeObject
BoundMethod_Type =
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "bound method",
  sizeof (BoundMethodObject),
  0,
  (destructor)BoundMethod_Dealloc,   /*tp_dealloc*/
  0,				     /*tp_print*/
  (getattrfunc)BoundMethod_GetAttr,  /*tp_getattr*/
  0,				     /*tp_setattr*/
  (cmpfunc)BoundMethod_Compare,	     /*tp_compare*/
  (reprfunc)BoundMethod_Repr,	     /*tp_repr*/
  0,				     /*tp_as_number*/
  0,				     /*tp_as_sequence*/
  0,				     /*tp_as_mapping*/
  0,				     /*tp_hash*/
  (ternaryfunc)BoundMethod_Call,     /*tp_call*/
};


PyObject *
BoundMethod_New(func, self)
	PyObject *func;
	PyObject *self;
{
  BoundMethodObject *im;

  if (! PyCallable_Check(func))
    {
      PyErr_BadInternalCall();
      return NULL;
    }
  
  im = PyObject_NEW(BoundMethodObject, &BoundMethod_Type);
  if (im == NULL)
    return NULL;
  
  Py_INCREF(func);
  im->bm_func = func;
  
  Py_XINCREF(self);
  im->bm_env = self;

  return (PyObject *)im;
}



static char Object_bind_Doc[] =
"Create a bound method object from (FUNCTION, INSTANCE).";

static PyObject *
Object_bind(unused, args)
     PyObject *unused;
     PyObject *args;
{
  PyObject *func;
  PyObject *self;

  if (!PyArg_ParseTuple(args, "OO", &func, &self))
    return NULL;
  
  return BoundMethod_New(func, self);
}




static struct PyMethodDef
Object_Methods[] =
{
  {"bind", (PyCFunction)Object_bind,	1, Object_bind_Doc},
  {NULL},		/* sentinel */
};


char Object_Doc[] =
"Alternate class system implemented using a meta class";


void
initobject()
{
  PyObject *m;
  PyObject *co;

  m = Py_InitModule4("object", Object_Methods, Object_Doc, (PyObject *)NULL,
                     PYTHON_API_VERSION);

  /*
   *  MetaType is derived from Type and overrides its tp_call.  It has
   *  a single instance: MetaClass.
   */  
  memcpy(&MetaType_Type, &PyType_Type, sizeof PyType_Type);
  MetaType_Type.tp_name = "object.type";
  MetaType_Type.tp_call = (ternaryfunc)MetaType_Call;

  /*
   * MetaClass is derived from Class
   */
  memcpy(&MetaClass_Type, &PyClass_Type, sizeof PyClass_Type);
  MetaClass_Type.ob_type = &MetaType_Type;
  MetaClass_Type.tp_name = "object.class";
  MetaClass_Type.tp_basicsize = sizeof (MetaClassObject);
  MetaClass_Type.tp_getattro = (getattrofunc) MetaClass_GetAttr;
  MetaClass_Type.tp_setattro = (setattrofunc) MetaClass_SetAttr;
  MetaClass_Type.tp_dealloc = (destructor) MetaClass_Dealloc;
  MetaClass_Type.tp_repr = (reprfunc) MetaClass_Repr;
  MetaClass_Type.tp_call = (ternaryfunc) MetaClass_Call;

  /*
   * Bootstrap the object system by creating an instance of the
   * MetaClass_Type and installing it in the module dictionary.
   * It is *the* meta class object.  It has no base class and is an
   * instance of itself.
   */
  MetaClass = MetaClass_New(NULL, NULL, NULL,
                            PyString_FromString("object.meta_class"));
  ((MetaClassObject*)MetaClass)->mc.instantiate = MetaClass_Instantiate;
  ((MetaClassObject*)MetaClass)->mc.class = (MetaClassObject*)MetaClass;
  MetaClass_InstallMethods(MetaClass, MetaClass_Methods);

  PyObject_SetAttrString(m, "meta_class", MetaClass);

  /*
   * ...and then sub-class it, overriding the instantiation method so
   * that it creates new classes, not new meta classes.
   */
  BaseClass = MetaClass_New(MetaClass, Py_BuildValue("(O)", MetaClass),
                            NULL, PyString_FromString("object.base_class"));
  ((MetaClassObject*)BaseClass)->mc.instantiate = BaseClass_Instantiate;
  MetaClass_InstallMethods(BaseClass, BaseClass_Methods);
  PyObject_SetAttrString(m, "base_class", BaseClass);

  /*
   * For convenience, create an instance of the base_class
   */
  co = MetaClass_New(BaseClass, NULL, NULL, PyString_FromString("object.base"));
  MetaClass_InstallMethods(co, Base_Methods);
  PyObject_SetAttrString(m, "base", co);
}

